<?php

declare(strict_types=1);

namespace BitBag\SyliusBonusPointsPlugin\EventListener;

use BitBag\SyliusBonusPointsPlugin\Calculator\DelegatingBonusPointsStrategyCalculatorInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\BonusPointsInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\BonusPointsStrategyInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\CustomerBonusPointsInterface;
use BitBag\SyliusBonusPointsPlugin\Repository\BonusPointsStrategyRepositoryInterface;
use DateTime;
use Doctrine\ORM\EntityManagerInterface;
use Sylius\Component\Core\Model\CustomerInterface;
use Sylius\Component\Resource\Factory\FactoryInterface;
use Sylius\Component\Resource\Repository\RepositoryInterface;
use Symfony\Component\EventDispatcher\GenericEvent;

final class BonusPointsEventListener
{
    private $bonusPointsStrategyRepository;
    /**
     * @var DelegatingBonusPointsStrategyCalculatorInterface
     */
    private $delegatingBonusPointsStrategyCalculator;
    /**
     * @var FactoryInterface
     */
    private $bonusPointsFactory;
    /**
     * @var EntityManagerInterface
     */
    private $bonusPointsEntityManager;
    /**
     * @var RepositoryInterface
     */
    private $customerBonusPointsRepository;
    /**
     * @var FactoryInterface
     */
    private $customerBonusPointsFactory;

    public function __construct(
        BonusPointsStrategyRepositoryInterface $bonusPointsStrategyRepository,
        DelegatingBonusPointsStrategyCalculatorInterface $delegatingBonusPointsStrategyCalculator,
        FactoryInterface $bonusPointsFactory,
        EntityManagerInterface $bonusPointsEntityManager,
        RepositoryInterface $customerBonusPointsRepository,
        FactoryInterface $customerBonusPointsFactory
    ) {
        $this->bonusPointsStrategyRepository = $bonusPointsStrategyRepository;
        $this->delegatingBonusPointsStrategyCalculator = $delegatingBonusPointsStrategyCalculator;
        $this->bonusPointsFactory = $bonusPointsFactory;
        $this->bonusPointsEntityManager = $bonusPointsEntityManager;
        $this->customerBonusPointsRepository = $customerBonusPointsRepository;
        $this->customerBonusPointsFactory = $customerBonusPointsFactory;
    }

    public function __invoke(GenericEvent $event, string $eventName)
    {
        $subject = $event->getSubject();

        /** @var BonusPointsStrategyInterface[] $bonusPointsStrategies */
        $bonusPointsStrategies = $this->bonusPointsStrategyRepository->findBy(['event' => $eventName]);

        $bonusPointsTotal = 0;
        foreach ($bonusPointsStrategies as $bonusPointsStrategy) {
            $bonusPointsTotal += $this->delegatingBonusPointsStrategyCalculator->calculate(
                $subject,
                $bonusPointsStrategy,
                0
            );
        }

        if ($bonusPointsTotal === 0) {
            return;
        }

        /** @var BonusPointsInterface $bonusPoints */
        $bonusPoints = $this->bonusPointsFactory->createNew();

        $bonusPoints->setPoints($bonusPointsTotal);
        $bonusPoints->setIsUsed(false);
        $bonusPoints->setExpiresAt((new DateTime())->modify('+1 year'));

        $this->getCustomerBonusPoints($subject)->addBonusPoints($bonusPoints);

        $this->bonusPointsEntityManager->persist($bonusPoints);
        $this->bonusPointsEntityManager->flush();
    }

    private function getCustomerBonusPoints(CustomerInterface $customer): CustomerBonusPointsInterface
    {
        /** @var CustomerBonusPointsInterface $customerBonusPoints */
        $customerBonusPoints = $this->customerBonusPointsRepository->findOneBy([
                'customer' => $customer,
            ]) ?? $this->customerBonusPointsFactory->createNew();

        if (null === $customerBonusPoints->getId()) {
            $customerBonusPoints->setCustomer($customer);

            $this->customerBonusPointsRepository->add($customerBonusPoints);
        }

        return $customerBonusPoints;
    }
}
